/*
 * ShellFish 
 * =========
 * GCC ISO C99 style c source for Raspberry Pi
 * Writen by M.F. Blokker - ferenc@sanghalung.com
 *
 * This program comes with ABSOLUTELY NO WARRANTY, to the extent
 * permitted by applicable law.
 *
 * Description
 * ========================================================================================
 * ShellFish, Run scripts on input / output events of a Velleman P8055 [PIC16C745] usb card
 *
 * Hardware Input / Output triggers a system shell supporting the input and output state as 
 * formated and easy to use parameters in your shell script. Create easy bash and php/html 
 * scripts to control your hardware without the hassle of programming in the c language but 
 * still want to enjoy the rappid interactions of a c program.
 *
 * ShellFish includes a event driven bash script and a php script illustrating how to use
 * the program. They are merely examples, review the code before operational use.
 *
 */

// lib headers

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <usb.h>
#include <assert.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#include <sys/select.h>
#include <stropts.h>
#include <signal.h>

// headers

#include "libk8055.h"
#include "shellfish.h"

// Const Config

const int CONFIG_OK = 0;
const int CONFIG_HELP_SHOWN = 1;
const int CONFIG_NO_PERMISSION_RUN = 2;
const int CONFIG_NO_PERMISSION_READ = 3;
const int CONFIG_RUNFILE_NOT_FOUND = 4;
const int CONFIG_COULD_NOT_OPEN_BOARD = 5;

// Const Events

const int EV_CARD_CONNECT 				= 1;
const int EV_CARD_DISCONNECT 				= 2;
const int EV_APP_START 					= 3;
const int EV_APP_EXIT 					= 4;
const int EV_APP_REFRESH	 			= 5;
const int EV_BUTTON_SIG_UP 				= 6;
const int EV_BUTTON_SIG_DOWN 				= 7;
const int EV_BUTTON_SIG_DOWN_AFTER_HOLD 		= 8;
const int EV_ANALOG1_VAL_CHANGED 			= 9;
const int EV_ANALOG2_VAL_CHANGED 			= 10;
const int EV_COUNTER1_VAL_CHANGED 			= 11;
const int EV_COUNTER2_VAL_CHANGED 			= 12;
const int EV_KEYBOARD_INPUT 				= 13;
const int EV_TIMER 					= 14;
const int EV_FUNCTIONKEY 				= 15;

// Const Event Names

const char EV_CARD_CONNECT_NAME[3] 			= "CC";
const char EV_CARD_DISCONNECT_NAME[3] 			= "CD";
const char EV_APP_START_NAME[3] 			= "ST";
const char EV_APP_EXIT_NAME[3] 				= "EX";
const char EV_APP_REFRESH_NAME[3] 			= "RF";
const char EV_BUTTON_SIG_UP_NAME[3] 			= "UP";
const char EV_BUTTON_SIG_DOWN_NAME[3] 			= "DN";
const char EV_BUTTON_SIG_DOWN_AFTER_HOLD_NAME[3] 	= "DH";
const char EV_ANALOG1_VAL_CHANGED_NAME[3] 		= "A1";
const char EV_ANALOG2_VAL_CHANGED_NAME[3] 		= "A2";
const char EV_COUNTER1_VAL_CHANGED_NAME[3] 		= "C1";
const char EV_COUNTER2_VAL_CHANGED_NAME[3] 		= "C2";
const char EV_KEYBOARD_INPUT_NAME[3] 			= "KB";
const char EV_TIMER_NAME[3] 				= "TM";
const char EV_FUNCTIONKEY_NAME[3] 			= "FN";

// handler

volatile sig_atomic_t Terminate_ShellFish;

// vars int

int Usb_Board_Address=0;				// P8055 Interace card 0,1,2 or 4
int Mod_Number=1;					// Default mod number
int Quiet_Mode=0;					// Mute certain text output
int Input_Debounce=100;					// Input 1..5 debounce time
int Counter_Debounce=35;				// Counter 1..2 debounce time
int Do_HTTP_update=0;					// if nonzero the status will be passed to localhost www service
int Do_Alive_Blinker=0;					// if nonzero the alive signal will be enabled
int Do_HTTP_Blinker=0;					// if nonzero the update signal will be enabled
int Do_BZY_Blinker=0;					// if nonzero the BZY signal will be enabled
int Run_File_Type=0;					// Executable file type
int Logger_Mode=0;					// Errors must be written to stderr

// vars char

char Program_Version[25]="0.0.1";			// Version of ShellFish
char Programmer[30]="Marcel F. Blokker";		// Programmer name
char Programmer_Email[30]="ferenc@sanghalung.com";	// Programmer email address
char Url_Server_Stat_Update[1024]="http://localhost/shellfish_status_collector.php";	// Default Url where updates will be sent / retrieved
char Run_File[1024]="./shellfish.sh";			// Name of executable file
char Shell_Data_Folder[1024]=".";			// Shell data folder 

// var char colors

char colGREEN[12]="\e[92m";				// Green
char colNORMAL[12]="\e[0m";				// Normal terminal color
char colBLUE[12]="\e[94m";				// Blue
char colLBLUE[12]="\e[36m";				// Light blue
char colRED[12]="\e[31m";				// Red
char colYEL[12]="\e[33m";

/*
 * Signal handler 
 *
 */
 
void Sig_Handler(int signum)
{
	Terminate_ShellFish = 1;			// If called by system, set the terminate flag
}

/*
 * Main (header text)
 *
 */

void  Print_Header(void)
{
	if(Quiet_Mode==0){
		
		printf("\n%sShellFish VELLEMAN P8055 / VM110 controller for LINUX version %s\n\n%s"
			, colYEL, Program_Version, colNORMAL);
		
		printf("%sCopyrights 2021 by %s - %s.\n%s", colBLUE, Programmer, Programmer_Email, colNORMAL);
		printf("%sThis is free software; see the source for copying conditions.\n%s", colBLUE, colNORMAL);
		printf("%sThis program comes with ABSOLUTELY NO WARRANTY.%s\n\n", colBLUE, colNORMAL);
	}
	
}

/*
 * Main (help text) 
 *
 */

void Print_Help(void)
{
	if(Quiet_Mode==0) {
		printf("%s", colGREEN);
		printf("Help\n\n");
		printf("Parameter         Description\n\n");
		printf("--help            Show this help text.\n");
		printf(" -a(Number)  0..1 Enable heartbeat signal on Output 8. Default is 0.\n");
		printf(" -b(Number)  0..1 Enable update signal on Output 7. Default is 0.\n");
		printf(" -c(Number)  0..1 Enable busy signal on Output 6. Default is 0.\n");
		printf(" -C(URL)     Text URL of status collector.\n");
		printf(" -d(Name)    Text (Path) Path of shell data folder.\n");
		printf(" -i(Number)  0..3 Interface number of P8055 / VM110 usb card. Default is 0.\n");	
		printf(" -l(Number)  0..1 Log errors to stderr. Default is 0.\n");
		printf(" -m(Mod)     0..N Module number passed to event script. Default is 0.\n");
		printf(" -s(Number)  0..1 Silence output to screen. Default is 0.\n");
		printf(" -u(Number)  0..1 Enable update status collector to web service. Default is 0.\n");	
		printf("                 URL of status collector required (see -C)\n");
		printf("                 Refer to the documentation for URL parameters description.\n");
		printf(" -x(Name)    Text ((Path)Filename) of event script.\n");
		printf("                 Default is ./shellfish.sh.\n");
		printf("                 Refer to the documentation how to setup event scripts.\n");
		printf(" -X(Number)  0..N Type of command file. Default is 0 (bash shell).\n");
		printf("\n");
		printf("%s", colNORMAL);
	}
}

/*
 * Main (code)
 *
 */

int main(int argc,char *params[])
{
	
	int Ret_Stat=0;
	char Current_Path[1024];

	// init
	Usb_Board_Address=0;	// Default interface
	Mod_Number=0;		// Default module
	
	// read command line parameters
	Ret_Stat=Get_Cmd_Params(argc, params);

	if((Ret_Stat>CONFIG_HELP_SHOWN) || (Ret_Stat==CONFIG_OK)){
		// Main header
		Print_Header();
	}
	
	if(Ret_Stat==CONFIG_NO_PERMISSION_RUN){
		if(Quiet_Mode==0) printf("No permission to run file %s.\n", Run_File);
		if(Logger_Mode==1) fprintf(stderr, "No permission to run file %s.\n", Run_File);
	}else if(Ret_Stat==CONFIG_NO_PERMISSION_READ){
		if(Quiet_Mode==0) printf("No permission to read from file %s.\n", Run_File);
		if(Logger_Mode==1) fprintf(stderr, "No permission to read from file %s.\n", Run_File);
	}else if(Ret_Stat==CONFIG_RUNFILE_NOT_FOUND){
		if(Quiet_Mode==0) printf("File %s doesn't exist.\nCheck or specify the -x parameter.\n", Run_File);
		if(Logger_Mode==1) fprintf(stderr, "File %s doesn't exist.\nCheck or specify the -x parameter.\n", Run_File);
	}

	// Config errors
	if (Ret_Stat>CONFIG_HELP_SHOWN){
		if(Quiet_Mode==0) printf("\nCan't continue due to configuration failure (error %i).\n", Ret_Stat);
		if(Logger_Mode==1) fprintf(stderr, "\nCan't continue due to configuration failure (error %i).\n", Ret_Stat);
		return Ret_Stat;
	// Help was shown
	}else if(Ret_Stat==CONFIG_HELP_SHOWN){
		return Ret_Stat;
	}

	if(Logger_Mode==1) fprintf(stderr, "Started\n");
	
	// Get the execution folder
	if(getcwd(Current_Path,sizeof(Current_Path)) != NULL){
		// Report the execution folder
		if(Quiet_Mode==0) printf("Current execution path: %s\n", Current_Path);
	}

	// Open interface
	if ( OpenDevice(Usb_Board_Address)<0 ) {
		
		// Could not open
		if(Quiet_Mode==0) printf("%sCould not open K8055 interface %d on USB (error 99).\n",colNORMAL, Usb_Board_Address);
		if(Logger_Mode==1) fprintf(stderr, "%sCould not open K8055 interface %d on USB (error 99).\n",colNORMAL, Usb_Board_Address);
		return CONFIG_COULD_NOT_OPEN_BOARD;
			
	}else{
		// Interface is open, call mods
		
		// Start
		ShellFish(Usb_Board_Address);
		
		// close card
		CloseDevice();
	}
	
	if(Quiet_Mode==0) printf("%sBye!\n", colNORMAL);
	if(Logger_Mode==1) fprintf(stderr, "%sBye!\n", colNORMAL);
	return 0;

}

/*
 * Get_Cmd_Params : read paramters from commandline
 *
 */

int Get_Cmd_Params(int argc,char *params[])
{

	int i;

	for (i=1; i<argc; i++){

		// Option --help
		if (!strncmp(params[i],"--help",6)){

			// Main header
			Print_Header();
			// Print help
			Print_Help();
			return CONFIG_HELP_SHOWN;
	
		}
		
		// Option --hdr
		if (!strncmp(params[i],"--hdr",5)){

			// Main header
			Print_Header();
	
		}

		// Option a - Alive status
		if (!strncmp(params[i],"-a",2)){

			Cv_String_To_Int(params[i]+2, &Do_Alive_Blinker );

		}

		// Option b - Update signal
		if (!strncmp(params[i],"-b",2)){

			Cv_String_To_Int(params[i]+2, &Do_HTTP_Blinker );

		}

		// Option c - BZY signal
		if (!strncmp(params[i],"-c",2)){

			Cv_String_To_Int(params[i]+2, &Do_BZY_Blinker );

		}

		// Option C - url of status collector
		if (!strncmp(params[i],"-C",2)){

			strcpy(Url_Server_Stat_Update, params[i]+2);
	
		}
		
		// Option d - Shell data folder
		if (!strncmp(params[i],"-d",2)){

			strcpy(Shell_Data_Folder, params[i]+2);
	
		}
		

		// Option i - Interface 0,1,2 or 3
		if (!strncmp(params[i],"-i",2)){

			Cv_String_To_Int(params[i]+2, &Usb_Board_Address );

		}
		
		// Option m - Mod number
		if (!strncmp(params[i],"-l",2)){

			Cv_String_To_Int(params[i]+2, &Logger_Mode );
	
	
		}

		// Option m - Mod number
		if (!strncmp(params[i],"-m",2)){

			Cv_String_To_Int(params[i]+2, &Mod_Number );
	
	
		}
		
		// Option s - Silenced output
		if (!strncmp(params[i],"-s",2)){

			Cv_String_To_Int(params[i]+2, &Quiet_Mode );
	
	
		}
		
		// Option u - Network update
		if (!strncmp(params[i],"-u",2)){

			Cv_String_To_Int(params[i]+2, &Do_HTTP_update );
	
	
		}
		
		// Option X - x file TYPE
		if (!strncmp(params[i],"-X",2)){

			Cv_String_To_Int(params[i]+2, &Run_File_Type );
	
	
		}

		// Option x - Name and Path Script file
		if (!strncmp(params[i],"-x",2)){

			strcpy(Run_File, params[i]+2);
	
		}


	}
	
	// Check bash file
	
	if( access( Run_File, F_OK ) != -1 ) {
		// file exists
		if( access( Run_File, R_OK ) != -1 ) {
			// file readable
			if( access( Run_File, X_OK ) != -1 ) {
				// file readable
			} else {
				
				return CONFIG_NO_PERMISSION_RUN;
			}
		} else {
			return CONFIG_NO_PERMISSION_READ;
		}
	} else {
		return CONFIG_RUNFILE_NOT_FOUND;
	}
	
	return 0;
}

/*
 * ShellFish: Run scripts on hardware input events
 *
 */

void ShellFish()
{
	// vars
	struct timeval 
		Start_Time, End_Time;		// start end endtime for elapsed time calcs
	int i=0;				// loop variable
	long State_Buttons=0;			// button states

	long State_Analog_1=0;			// analog 1
	long State_Analog_2=0;			// analog 2

	int State_SimAnalogIn_1=0;		// analog 1 simulated
	int State_SimAnalogIn_2=0;		// analog 2 simulated
	int SimAnalogIn_1=0;			// analog 1 simulation on
	int SimAnalogIn_2=0;			// analog 2 simulation on

	long Counter_1=0; 			// counter 1
	long Counter_2=0;			// counter 2
	int Sub_Counter_1=0; 			// counter 1
	int Sub_Counter_2=0;			// counter 2

	int Out_State=0;			// state of the output gates
	int Dac_Out_State_1=0;			// state of the analog output 1
	int Dac_Out_State_2=0;			// state of the analog output 2
	int Button_State=0;			// state of the buttons

	long Prev_State_Analog_1=0;		// Previous analog 1
	long Prev_State_Analog_2=0;		// Previous analog 2
	int Prev_State_SimAnalogIn_1=0;		// analog 1 simulated
	int Prev_State_SimAnalogIn_2=0;		// analog 2 simulated
	long Prev_Counter_1=0; 			// Previous counter 1
	long Prev_Counter_2=0;			// Previous counter 2
	int Prev_Out_State=-0;			// Previous output state
	int Prev_Button_State=-0;		// Previous button state
	int Prev_Dac_Out_State_1=0;		// state of the analog output 1
	int Prev_Dac_Out_State_2=0;		// state of the analog output 2

	int Bit_Offset[9] = {0};		// bit poistion offsets (1,2,4,8,16,32,128)
	int Wait_For_Release[9] = {0};		// wait for released button mode
	int Button_Hold[9] = {0};		// Indicate a hold button
	int Button_Hold_Release[9] = {0};	// Indicate a hold released button mode
	int Queued_Button[9] = {0};		// if set to 1 a button press is emulated
	double Elapsed_Time;			// Time already elapsed from the sampled time
	double Time_Slack=0;			// time exceed maximum interval
	int ShellFish_Started=1;		// Indicates a succesfull start
	int Bit_Offset_Step=1;			// used to calc bit offsets
	int Update_Step=0;			// Current counter step for update server
	int Update_Start=3;			// Exec after steps proceeded 
	int Alive_Step=0;			// Current counter alive step for alive signal
	int Alive_Start=6;			// Exec after steps proceeded
	int Alive_State_Toggle=0;		// Alive signal if off
	int Return_Status=0;			// return from usb read
	int Connection_Error_Printed=0;		// error printed on screen
	int Delta_Step=0;			// Number of .5 second steps from delta time
	int Function_Key[256] = {0};		// Function keys
	int Timers[256] = {0};			// Timer countdown cueue


	// Init Bit_Offset
        for (i=1;i<9;i++) {
		Bit_Offset[i]=Bit_Offset_Step;
		Bit_Offset_Step=Bit_Offset_Step*2;
        }

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(0).\n");

	// reset digital output
	WriteAllDigital(0);
	
	// reset analog output
	
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:OutputAnalogChannel(1,0).\n");
	
	OutputAnalogChannel(1, 0);

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:OutputAnalogChannel(2,0).\n");

	OutputAnalogChannel(2, 0);

	// Reset counters

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:ResetCounter(1).\n");
	ResetCounter(1);

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:ResetCounter(2).\n");
	ResetCounter(2);
	
	// Set the debounce
	// This will prevent counters to increase on verry small intervals
	// (as happens when you jiggle the test buttons on the board when in hold state)
	SetCounterDebounceTime(1,Counter_Debounce);
	SetCounterDebounceTime(2,Counter_Debounce);

	// mark start
	gettimeofday(&Start_Time, NULL);

	// main loop
	while(-1){
	
/*
 * Read all status and reconnect if yielding error
 *
 */
		// ReadAllValues is not logged unless an error appears, it would explode the log otherwise
 
		// get the state of the usb card
		Return_Status=ReadAllValues(&State_Buttons, &State_Analog_1, &State_Analog_2, &Counter_1, &Counter_2);
		
		// check for errors and reopen, the usb card may be disconnected
		if (Return_Status!=0){
			
			// Log the error
			if(Logger_Mode==1) fprintf(stderr, "libk8055:ReadAllValues returned an error.\n");
		
			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:CloseDevice.\n");

			// close the device
			CloseDevice();
			
			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:usb_find_busses.\n");

			// iterate avaiable busses
		        usb_find_busses();
			
			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:usb_find_devices.\n");

			// iterate devices
		        usb_find_devices();
			
			// wait for a while, this shouldn't be [...]			
			usleep(500000);

			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:OpenDevice.\n");
			
			// re-open interface			
			if ( OpenDevice(Usb_Board_Address)<0 ) {
				
				// Log the error
				if(Logger_Mode==1) fprintf(stderr, "libk8055:OpenDevice returned an error.\n");
				
				// if no error was already printed on screen
				if (Connection_Error_Printed==0){

					// Reset values
					Out_State=0;
					Button_State=0;
					Counter_1=0;
					Counter_2=0;
					Sub_Counter_1=0;
					Sub_Counter_2=0;
					State_Analog_1=0;
					State_Analog_2=0;
					Dac_Out_State_1=0;
					Dac_Out_State_2=0;

					// print error to screen
					if(Quiet_Mode==0) printf("Could not read from K8055 usb interface.\n");
					if(Quiet_Mode==0) printf("Please check your hardware and cable.\n");
					
					// Log
					if(Logger_Mode==1) fprintf(stderr, "Could not read from K8055 usb interface.\n");
					
					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_CARD_DISCONNECT, Out_State, 0, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					
					// do not drop this message again
					Connection_Error_Printed=1;
				}
			}else{
				
				// device is open
				if(Quiet_Mode==0) printf("Reconnected with K8055 usb interface.\n");
				// log
				if(Logger_Mode==1) fprintf(stderr, "Reconnected with K8055 usb interface.\n");

				// Shell and process shell return
				Out_State=ShellFish_Run_Shell(EV_CARD_CONNECT, Out_State, 0, Button_State, Delta_Step, 
				                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
							      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
							      Queued_Button, Button_Hold, Function_Key, Timers,
							      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);

				// Make sure future connection errors are printend on the screen again
				Connection_Error_Printed=0;
			}
		}

/*
 * Button actions
 *
 */
 
		// Check all 8 buttons
		for (i=1;i<9;i++) {

			// get the buttons pressed / depressed state
			int btn=0;
			btn=((State_Buttons & Bit_Offset[i]) == Bit_Offset[i]);
			// Hardware button is pressed, cancel hold
			if (btn==1) Button_Hold[i]=0;

			// Detection of counter increase while there was not button press detected
			// This happens when code is bzy doing something and button 1 or 2 are
			// pressed for a verry short time. If the counter was increased the
			// button press will be simulated. This only works for button 1 and 2
			
			// button 1 was not pressed
			if((i==1) & (btn==0)){
				// But counter 1 was increased, evaluation of the button press was too late
				if((Counter_1-Prev_Counter_1)>0){
					// Buttons was not in hold state
					if (Button_Hold[i]==0){
						// Hit the button
						btn=1;
					}
				}
			}

			// button 2 was not pressed
			if((i==2) & (btn==0)){
				// But counter 2 was increased, evaluation of the button press was too late
				if((Counter_2-Prev_Counter_2)>0){
					// Buttons was not in hold state
					if (Button_Hold[i]==0){
						// Hit the button
						btn=1;
					}
				}
			}
			
			// if the button is pressed or a pending command is active
			if (btn==1 || Queued_Button[i]==1 || Button_Hold[i]==1 ){
				// Merge the button state 
				Button_State=Merge_Bit(Button_State,i,1);

				// if button is not released, aka still pressed
				if (Wait_For_Release[i]==0){
					
					// If button 1 was softwarematically 
					if((i==1) & (Queued_Button[i]==1)){
						Sub_Counter_1++;
					}
					// If button 2 was softwarematically 
					if((i==2) & (Queued_Button[i]==1)){
						Sub_Counter_2++;
					}
				}

				// no more pending command
				Queued_Button[i]=0;

				// if button is not released, aka still pressed
				if (Wait_For_Release[i]==0){
					
					// Softwarematically hold
					Button_Hold_Release[i]=0;
					if (Button_Hold[i]==1){
						Button_Hold_Release[i]=1;
					}

					// Turn the bzy button off
					if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,1);
					
					// Log message
					if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);
										
					// write output to usb hardware
					WriteAllDigital(Out_State);
					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					// button in release mode, aka no actions until button is released
					Wait_For_Release[i]=1;
					
					// wait a while to prevent unwanted counter flipping
					usleep(200000);
					
					// Turn the bzy button off
					if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,0);
					
					// Log message
					if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);
					
					// write output to usb hardware
					WriteAllDigital(Out_State);
					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_BUTTON_SIG_UP, Out_State, i, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					
				}
			}
	
			// if the button is not pressed
			if ((btn==0) & (Button_Hold[i]==0)){
				// Merge button bit into place
				Button_State=Merge_Bit(Button_State,i,0);
				
				// Release detected					
				if (Wait_For_Release[i]==1){
					
					// Turn the bzy button on
					if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,1);
					
					// Log message
					if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);
					
					// write output to usb hardware
					WriteAllDigital(Out_State);
					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					// Turn the bzy button off
					if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,0);

					// Log message
					if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);

					// write to usb hardware
					WriteAllDigital(Out_State);					
					// Button was released after hold
					if(Button_Hold_Release[i]==1){

						// wait a while to prevent unwanted counter flipping
						usleep(200000);
						
						// Turn the button hold release off
						Button_Hold_Release[i]=0;
						// Shell and process shell return
						Out_State=ShellFish_Run_Shell(EV_BUTTON_SIG_DOWN_AFTER_HOLD, Out_State, i, Button_State, Delta_Step, 
						                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
									      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
									      Queued_Button, Button_Hold, Function_Key, Timers,
									      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);

					}else{

						// wait a while to prevent unwanted counter flipping
						usleep(200000);
						
						// Shell and process shell return
						Out_State=ShellFish_Run_Shell(EV_BUTTON_SIG_DOWN, Out_State, i, Button_State, Delta_Step, 
						                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
									      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
									      Queued_Button, Button_Hold, Function_Key, Timers,
									      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					}

					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);

				}
				
				// done waiting for release of the button
				Wait_For_Release[i]=0;

			}
			
		}
		
		if(ShellFish_Started==1){
			// Shell and process shell return
			Out_State=ShellFish_Run_Shell(EV_APP_START, Out_State, 0, Button_State, Delta_Step, 
			                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
						      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
						      Queued_Button, Button_Hold, Function_Key, Timers,
						      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);

			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);

			// write to usb hardware
			WriteAllDigital(Out_State);
			// Set nul, so it runs only once
			ShellFish_Started=0;
			
			// Shell and process shell return
			Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
						      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
						      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
						      Queued_Button, Button_Hold, Function_Key, Timers,
						      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
			
		}
		
		// Loop trough function key array
		for(i=1; i<256; i++){
			// if element is set
			if(Function_Key[i]==1){
				// Turn it off
				Function_Key[i]=0;
				// Shell and process shell return
				Out_State=ShellFish_Run_Shell(EV_FUNCTIONKEY, Out_State, i, Button_State, Delta_Step, 
							      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
							      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
							      Queued_Button, Button_Hold, Function_Key, Timers,
							      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
								
			}
		}
		
		

/*
 * Keyboard commands
 *
 */
		// If keyboard is hit
		if (Keyboard_Hit()==1){
			
			// get the char
			char c = Get_Keyboard();

			// exit
			if (c == 'x'){
				break;
			}else{
				// Not CR, not LF, not -1 / 255
				if((c!=10) & (c!=13) & (c!=-1) & (c!=255)){
					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_KEYBOARD_INPUT, Out_State, c, Button_State, Delta_Step, 
								      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);

				}
			}
		}
	
/*
 * Sequence stepper, Measures elapsed time and triggers events ofter designated expirations 
 *
 */
	    	// get timer stop value
    		gettimeofday(&End_Time, NULL);
    		// compute elapsed time
    		Elapsed_Time = (End_Time.tv_sec - Start_Time.tv_sec) * 1000.0;      
    		Elapsed_Time += (End_Time.tv_usec - Start_Time.tv_usec) / 1000.0; 
		// Add exceeded time from prev interval
		Elapsed_Time += Time_Slack;

		// time elapsed (in units of 1/2 second)
		// Do not change the unit time unless Timers etc are adjusted too.
		if (Elapsed_Time > 500){
			
	        	// start timer
        		gettimeofday(&Start_Time, NULL);
			// Time exceeded 500
			Time_Slack = Elapsed_Time - 500;
			// Increase number of steps since start
			Delta_Step++;
/*
 * Timers expiration and event fire
 *
 */
			// All timers
			for(i=1;i<256;i++){
				// .5 Second untis left to be counted down
				if(Timers[i]>0){
					// Decrease unit
					Timers[i]--;
					// Timer reached zero
					if(Timers[i]==0){
					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_TIMER, Out_State, i, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
						
					}
				}
			}
			
/*
 * Determine is an update must be send to the webservice
 *
 */			
			
			// Increase update steps			
			Update_Step+=1;
			// Steps exceed update start
			if(Update_Step > Update_Start){
				// Reset update step for next start
				Update_Step=1;
				// if -u option is nonzero
				if (Do_HTTP_update!=0){
					// If the signal to Output 7 is allowed
	                                if(Do_HTTP_Blinker==1){
						// Set output 7 on
						Out_State=Merge_Bit(Out_State,7,1);
						Wait_For_Release[7]=0;

						// Log message
						if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);

						// write to usb hardware
						WriteAllDigital(Out_State);				
					}

					// Shell and process shell return
					Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
					                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
								      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
								      Queued_Button, Button_Hold, Function_Key, Timers,
								      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					// Shell to output status and getting return
					Out_State=ShellFish_HTTP_Update(Out_State, Button_State, Delta_Step, 
									Counter_1, Counter_2, State_Analog_1, State_Analog_2,
									&Dac_Out_State_1, &Dac_Out_State_2,Queued_Button, Button_Hold,
									Function_Key, Timers, &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
					// If the signal to Output 7 is allowed
	                                if(Do_HTTP_Blinker==1){
						// Set output 7 off
						Out_State=Merge_Bit(Out_State,7,0);
						// Turn wait for release off
						Wait_For_Release[7]=0;

						// Log message
						if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);

						// write to usb hardware
						WriteAllDigital(Out_State);
					}
				}
			}

/* 
 * Determine if the alive blinker must be set
 *
 */

			// Increase alive steps
			Alive_Step+=1;
			// Steps exceed alive start
			if(Alive_Step > Alive_Start){
				// reset step
				Alive_Step=1;				
				// if -a option is nonzero
				if (Do_Alive_Blinker!=0){
					// state is on
					if (Alive_State_Toggle==1){
						// turn off
						Alive_State_Toggle=0;
						// turn off the alive bit
		                                Out_State=Merge_Bit(Out_State,8,0);
						// do not wait for release
						Wait_For_Release[8]=0;
					// state is off
					}else{
						// state off, turn on
						Alive_State_Toggle=1;
						// turn on alive bit
		                                Out_State=Merge_Bit(Out_State,8,1);
						Wait_For_Release[8]=0;
					}
					

					// Log message
					if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);

					// write to usb hardware
					WriteAllDigital(Out_State);
				}			
			}
		}
		
/*
 * Check for changes and fire distinct event
 * 
 */		
		
		// If counter 1 has changed
		if(Prev_Counter_1!=Counter_1){
			
			// Shell and process shell return
			Out_State=ShellFish_Run_Shell(EV_COUNTER1_VAL_CHANGED, Out_State, 1, Button_State, Delta_Step, 
			                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
						      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
						      Queued_Button, Button_Hold, Function_Key, Timers,
						      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
			
		}
		
		// If counter 2 has changed
		if(Prev_Counter_2!=Counter_2){
			
			// Shell and process shell return
			Out_State=ShellFish_Run_Shell(EV_COUNTER2_VAL_CHANGED, Out_State, 2, Button_State, Delta_Step, 
			                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
						      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
						      Queued_Button, Button_Hold, Function_Key, Timers,
						      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
			
		}

		// If not in analog 1 simulation 
		if(SimAnalogIn_1==0){
		
			// If Analog 1 has changed
			if(Prev_State_Analog_1!=State_Analog_1){

				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,1);

				
				// Shell and process shell return
				Out_State=ShellFish_Run_Shell(EV_ANALOG1_VAL_CHANGED, Out_State, 1, Button_State, Delta_Step, 
							      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
							      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
							      Queued_Button, Button_Hold, Function_Key, Timers,
							      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
				
				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,0);
			}
		
		// In analog 1 simulation
		}else if(SimAnalogIn_1==1){

			// If Analog 1 simulated has changed
			if(Prev_State_SimAnalogIn_1!=State_SimAnalogIn_1){

				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,1);

				
				// Shell and process shell return
				Out_State=ShellFish_Run_Shell(EV_ANALOG1_VAL_CHANGED, Out_State, 1, Button_State, Delta_Step, 
							      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
							      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
							      Queued_Button, Button_Hold, Function_Key, Timers,
							      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
				
				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,0);
			}
			
		}

		// If not in analog 2 simulation 
		if(SimAnalogIn_2==0){
			
			// If Analog 2 has changed
			if(Prev_State_Analog_2!=State_Analog_2){
				
				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,1);

				// Shell and process shell return
				Out_State=ShellFish_Run_Shell(EV_ANALOG2_VAL_CHANGED, Out_State, 2, Button_State, Delta_Step, 
							      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
							      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
							      Queued_Button, Button_Hold, Function_Key, Timers,
							      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
				
				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,0);

			}

		// In analog 2 simulation
		}else if(SimAnalogIn_2==1){
			
			// If Analog 2 simulated has changed
			if(Prev_State_SimAnalogIn_2!=State_SimAnalogIn_2){

				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,1);

				
				// Shell and process shell return
				Out_State=ShellFish_Run_Shell(EV_ANALOG2_VAL_CHANGED, Out_State, 2, Button_State, Delta_Step, 
							      Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
							      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
							      Queued_Button, Button_Hold, Function_Key, Timers,
							      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
				
				// Turn the bzy button on
				if(Do_BZY_Blinker!=0) Out_State=Merge_Bit(Out_State,6,0);
			}
			
		}

		// If Dac 1 has changed
		if(Prev_Dac_Out_State_1!=Dac_Out_State_1){

			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:OutputAnalogChannel(1,%i).\n", Dac_Out_State_1);

			// Set Dac 1
			OutputAnalogChannel(1, Dac_Out_State_1);
		}

		// If Dac 2 has changed
		if(Prev_Dac_Out_State_2!=Dac_Out_State_2){


			// Log message
			if(Logger_Mode==1) fprintf(stderr, "libk8055:OutputAnalogChannel(2,%i).\n", Dac_Out_State_2);

			// Set Dac 2
			OutputAnalogChannel(2, Dac_Out_State_2);
		}

		// if the state has been changed of input / output / analoq / counters
		// note that the refresh will be also triggered when analog input changes when in simulation mode
		if(Prev_Button_State!=State_Buttons || Prev_Out_State!=Out_State || Prev_Counter_1!=Counter_1 
						    || Prev_Counter_2!=Counter_2 
						    || Prev_State_Analog_1!=State_Analog_1 
						    || Prev_State_Analog_2!=State_Analog_2
						    || Prev_State_SimAnalogIn_1!=State_SimAnalogIn_1 
						    || Prev_State_SimAnalogIn_2!=State_SimAnalogIn_2
						    || Prev_Dac_Out_State_1!=Dac_Out_State_1
						    || Prev_Dac_Out_State_2!=Dac_Out_State_2){

			// Shell and process shell return
			Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
			                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
						      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
						      Queued_Button, Button_Hold, Function_Key, Timers,
						      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
						      
		}

		// Set check vars to determine a future refresh event
		Prev_Button_State = State_Buttons;	// Set previous button state
		Prev_Out_State = Out_State;		// Set previous out state
		Prev_State_Analog_1=State_Analog_1;	// Set previous analog 1
		Prev_State_Analog_2=State_Analog_2;	// Set previous analog 2
		Prev_Counter_1=Counter_1; 		// Set previous counter 1
		Prev_Counter_2=Counter_2;		// Set previous counter 2
		Prev_Dac_Out_State_1=Dac_Out_State_1;	// Set previous Dac out 1
		Prev_Dac_Out_State_2=Dac_Out_State_2;	// Set previous Dac out 2
		Prev_State_SimAnalogIn_1=State_SimAnalogIn_1; // Set previous simulated analog 1
		Prev_State_SimAnalogIn_2=State_SimAnalogIn_2; // Set previous simulated analog 2
				
		// Check if use pressed control c and exit if so
		signal(SIGINT, Sig_Handler);		
		if(Terminate_ShellFish==1) break;
		
		// Check for terminate and exit if so
		signal(SIGTERM, Sig_Handler);		
		if(Terminate_ShellFish==1) break;
		
		// Check if use pressed control z and exit if so
		signal(SIGTSTP, Sig_Handler);		
		if(Terminate_ShellFish==1) break;

	}

/*
 * End of main loop
 *
 */

	
	// Reset states
	Button_State=0;
	Out_State=0;
	Dac_Out_State_1=0;
	Dac_Out_State_2=0;
	
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:OutputAnalogChannel(1,%i).\n", Dac_Out_State_1);
	
	OutputAnalogChannel(1, Dac_Out_State_1);
	
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:OutputAnalogChannel(2,%i).\n", Dac_Out_State_2);
	
	OutputAnalogChannel(2, Dac_Out_State_2);
	Counter_1=0;
	Counter_2=0;
	Sub_Counter_1=0;
	Sub_Counter_2=0;
	
	// Reset counters
	
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:ResetCounter(1).\n");

	ResetCounter(1);

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:ResetCounter(2).\n");

	ResetCounter(2);
	
	
	
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",Out_State);
	
	// write to usb hardware
	WriteAllDigital(Out_State);
	// Shell and process shell return
	Out_State=ShellFish_Run_Shell(EV_APP_REFRESH, Out_State, 0, Button_State, Delta_Step, 
	                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
				      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
				      Queued_Button, Button_Hold, Function_Key, Timers,
				      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);
	// Shell and process shell return
	Out_State=ShellFish_Run_Shell(EV_APP_EXIT, Out_State, 0, Button_State, Delta_Step, 
	                              Counter_1, &Sub_Counter_1, Counter_2, &Sub_Counter_2, 
				      State_Analog_1, State_Analog_2, &Dac_Out_State_1, &Dac_Out_State_2, 
				      Queued_Button, Button_Hold, Function_Key, Timers,
				      &State_SimAnalogIn_1, &State_SimAnalogIn_2, &SimAnalogIn_1, &SimAnalogIn_2);

}

/* 
 * ShellFish_HTTP_Update: Update status to webservice and process return commands
 *
 * Return commands, each separated with a space. Ex. "P2 A127 B255 O6 o1"
 *
 *	O[n]	- Raise digital output. (O1..O8)
 *	o[n]	- Lower digital output. (o1..o8)
 *	P[n]	- Raise/Lower digital input (P1..P5)
 *	I[n]	- Raise digital input (I1..I5)
 *	i[n]	- Lower digital input (i1..i5)
 *	A[v]	- Set voltage analog out A (A0..A255)
 * 	B[v]	- Set voltage analog out B (B0..B255)
 * 	Y[v]	- Simulate voltage in A (Y0..Y255)
 * 	y[v]	- Set simulation in A (y0..y1)
 *      Z[v]	- Simulate voltage in B (Z0..Z255)
 * 	z[v]	- Set simulation in B (z0..z1)
 *	R[n]	- Reset counter (R1..R2)
 * 	T[n]	- Set timer (T1=secs..T255=secs)
 * 	F[n]	- Raise function event (F0..F255)
 */
 
int ShellFish_HTTP_Update(int Current_Out_State, int Current_Button_State, int Current_Delta_Step, 
			  int Counter_1, int Counter_2, int State_Analog_1, int State_Analog_2, int *Dac_Out_State_1, int *Dac_Out_State_2,
			  int *Queued_Button, int *Button_Hold, int *Function_Key, int *Timers, 
			  int *State_SimAnalogIn_1, int *State_SimAnalogIn_2, int *SimAnalogIn_1, int *SimAnalogIn_2)
{
	
	char Response_Filename[1024];		// Filename of response file
	char Sys_Command[1024];			// System command
	char Bit_String_Out_State[20];		// Bit formatted strign of out state
	char Bit_String_Button_State[20];	// Bit formatted string of button state
	char File_Char;				// char read from file
	char Command_Char=' ';			// char of command number
	int Command_Int=0;			// converted command number
	int New_Out_State = Current_Out_State;  // Copy current state
	int Sys_Return=0;
	
	// Set the analog in value to either actual or simulated input
	int StateA1in=State_Analog_1;				// actual input 1
	int StateA2in=State_Analog_2;				// actual input 2
	if(*SimAnalogIn_1==1) StateA1in=*State_SimAnalogIn_1;	// Change to analog simulated 1
	if(*SimAnalogIn_2==1) StateA2in=*State_SimAnalogIn_2;	// Change to analog simulated 2
	
	// copy the outstate bit string
	strcpy(Bit_String_Out_State, Bit_Str(New_Out_State));
	// copy the button state bit string
	strcpy(Bit_String_Button_State, Bit_Str(Current_Button_State));
	// create filename
	snprintf(Response_Filename, 1024, "%s/wwwresponse.%i.%i.txt", Shell_Data_Folder, Mod_Number, Usb_Board_Address);
	// Check if response file exists
	if( access( Response_Filename, F_OK ) != -1 ) {
		// Delete it, its from a previous session
		if( unlink(Response_Filename)!=0){
			// Message
			if(Quiet_Mode==0) printf("Error deleting file %s\n", Response_Filename);
			// Log message
			if(Logger_Mode==1) fprintf(stderr, "Error deleting file %s\n", Response_Filename);
		}
	}
	// command wget
	snprintf(Sys_Command, 1024, "wget --no-check-certificate -q -O%s/wwwresponse.%i.%i.txt '%s?shellfishid=shellfishupd(%i,%i,%s,%s,%i,%i,%i,%i,%i,%i,%i,%i,%i)'", 
	Shell_Data_Folder, Mod_Number, Usb_Board_Address,
	Url_Server_Stat_Update, Mod_Number, Usb_Board_Address, Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step,
	Counter_1, Counter_2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "Start ShellFish_HTTP_Update:system(%s).\n", Sys_Command);
	// Run the shell
	Sys_Return=system(Sys_Command);
	// If failed
	if(Sys_Return == -1){
		// retrun with unmodified new state = current state
		return New_Out_State;
	}
	// Message
	if(Quiet_Mode==0) printf("Opening file: %s\n", Response_Filename);
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "Start ShellFish_HTTP_Update:system() Processing return data.\n");
	// Open the response file
	FILE *File_Handle = fopen(Response_Filename, "r");
	if(File_Handle != NULL) {
		// Endless loop
		while (-1){
			// Get one char
			File_Char = fgetc(File_Handle);
			// If eof, then break
			if (feof(File_Handle)) break ;
			// Output On command
			if (File_Char=='O'){
				// Get output number
				Command_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
				// Convert char to num
				Command_Int=Command_Char-48;
				// If the number is between 1 and 8
				if((Command_Int>0) & (Command_Int<9)){
					// Set de outstate
					New_Out_State=Merge_Bit(New_Out_State,Command_Int,1);
					// Message on screen
					if(Quiet_Mode==0) printf("Set output %i to ON\n", Command_Int);
				}
			// Output Off command
			}else if (File_Char=='o'){
				// Get output number
				Command_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
				// Convert char to num
				Command_Int=Command_Char-48;
				// If the number is between 1 and 8
				if((Command_Int>0) & (Command_Int<9)){
					// Set de outstate
					New_Out_State=Merge_Bit(New_Out_State,Command_Int,0);
					// Message on screen
					if(Quiet_Mode==0) printf("Set output %i to OFF\n", Command_Int);
				}
				
			
			// Button push/release command (emulate)
			}else if (File_Char=='P'){
				// Get output number
				Command_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
				// Convert char to num
				Command_Int=Command_Char-48;
				// If the number is between 1 and 8
				if((Command_Int>0) & (Command_Int<9)){
					// Set de outstate
					// New_Out_State=Merge_Bit(New_Out_State,Command_Int,0);
					// Message on screen
					Queued_Button[Command_Int]=1;
					Button_Hold[Command_Int]=0;
					// Message on screen
					if(Quiet_Mode==0) printf("Emulate button %i Push/Release\n", Command_Int);
				}
			// Button push command (emulate)
			}else if (File_Char=='I'){
				
				// Get output number
				Command_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
				// Convert char to num
				Command_Int=Command_Char-48;
				// If the number is between 1 and 8
				if((Command_Int>0) & (Command_Int<9)){
					// Set de outstate
					Queued_Button[Command_Int]=1;
					Button_Hold[Command_Int]=1;
					// Message on screen
					if(Quiet_Mode==0) printf("Emulate button %i Push/Release\n", Command_Int);
				}
			// Button release command (emulate)
			}else if (File_Char=='i'){
				
				// Get output number
				Command_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
				// Convert char to num
				Command_Int=Command_Char-48;
				// If the number is between 1 and 8
				if((Command_Int>0) & (Command_Int<9)){
					// Set de outstate
					Queued_Button[Command_Int]=1;
					Button_Hold[Command_Int]=0;
					// Message on screen
					if(Quiet_Mode==0) printf("Emulate button %i Push/Release\n", Command_Int);
				}
			// Reset counter
			}else if (File_Char=='R'){
				
				// Get output number
				Command_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
				// Convert char to num
				Command_Int=Command_Char-48;
				// If the number is between 1 and 8
				if((Command_Int>0) & (Command_Int<3)){
					
					// Log message
					if(Logger_Mode==1) fprintf(stderr, "libk8055:ResetCounter(%i).\n", Command_Int);
					
					// Set de outstate
					ResetCounter(Command_Int);
					// Message on screen
					if(Quiet_Mode==0) printf("Reset counter %i\n", Command_Int);	
				}
			// Analog A (1) out
			}else if (File_Char=='A'){
				// Value of new dac
				int Analog_Value_A;
				// Get the new value from file stream
				Analog_Value_A=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Analog_Value_A>-1) & (Analog_Value_A<256)){
					// Set the parameter
					*Dac_Out_State_1= Analog_Value_A;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Analog out A to %i\n", Analog_Value_A);
				}
			// Analog B (2) out
			}else if (File_Char=='B'){
				// Value of new dac
				int Analog_Value_B;
				// Get the new value from file stream
				Analog_Value_B=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Analog_Value_B>-1) & (Analog_Value_B<256)){
					// Set the parameter
					*Dac_Out_State_2= Analog_Value_B;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Analog out B to %i\n", Analog_Value_B);
				}

			// Sim Analog in Y (1) in VALUE
			}else if (File_Char=='Y'){
				// Value of new dac
				int Analog_Value_Y;
				// Get the new value from file stream
				Analog_Value_Y=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Analog_Value_Y>-1) & (Analog_Value_Y<256)){
					// Turno on sim
					*SimAnalogIn_1=1;
					// Set the parameter
					*State_SimAnalogIn_1= Analog_Value_Y;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Simulated Analog in Y to %i\n", Analog_Value_Y);
				}

			// Sim Analog in Y (1) on/off
			}else if (File_Char=='y'){
				// Value of new dac
				int Analog_Value_Y;
				// Get the new value from file stream
				Analog_Value_Y=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Analog_Value_Y>-1) & (Analog_Value_Y<2)){
					// Set the parameter
					*SimAnalogIn_1=Analog_Value_Y;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Simulation Analog in Y to %i\n", Analog_Value_Y);
				}


			// Sim Analog in Z (2) VALUE
			}else if (File_Char=='Z'){
				// Value of new dac
				int Analog_Value_Z;
				// Get the new value from file stream
				Analog_Value_Z=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Analog_Value_Z>-1) & (Analog_Value_Z<256)){
					// Turno on sim
					*SimAnalogIn_2=1;
					// Set the parameter
					*State_SimAnalogIn_2= Analog_Value_Z;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Simulated Analog in Z to %i\n", Analog_Value_Z);
				}
				
			// Sim Analog in Z (2) on/off
			}else if (File_Char=='z'){
				// Value of new dac
				int Analog_Value_Z;
				// Get the new value from file stream
				Analog_Value_Z=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Analog_Value_Z>-1) & (Analog_Value_Z<2)){
					// Set the parameter
					*SimAnalogIn_2=Analog_Value_Z;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Simulation Analog in Z to %i\n", Analog_Value_Z);
				}

			// Function
			}else if (File_Char=='F'){
				// Value of new dac
				int Function_Key_Value;
				// Get the new value from file stream
				Function_Key_Value=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Function_Key_Value>0) & (Function_Key_Value<256)){
					// Set the parameter
					Function_Key[Function_Key_Value]=1;
					// Message on screen
					if(Quiet_Mode==0) printf("Set Function call to %i\n", Function_Key_Value);
				}
			// Timer
			}else if (File_Char=='T'){
				// Value of new dac
				int Timer_Offset;
				int Timer_Value;
				// Get the new value from file stream
				Timer_Offset=fget_Numeric_Parm(File_Handle);
				// If value is 0-255
				if((Timer_Offset>0) & (Timer_Offset<256)){
					// Get '='
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Needs to be '='
					if(Command_Char=='='){
						// Get number of seconds
						Timer_Value=fget_Numeric_Parm(File_Handle);
						// Eq or less than 24 Hrs
						if((Timer_Value>-1) & (Timer_Value<1000000)){
							// Set the parameters
							Timers[Timer_Offset]=Timer_Value*2;
							// Message on screen
							if(Quiet_Mode==0) printf("Set Timer %i to %i\n", Timer_Offset, Timer_Value);
						}
					}
				}
			}
		}
		
		fclose(File_Handle);
		
	}else{
		
		if(Quiet_Mode==0) printf("Error reading: %s\n", Response_Filename);
		
		if(Logger_Mode==1) fprintf(stderr, "Error reading: %s\n", Response_Filename);
		
	}
	
	// Log message
	if(Logger_Mode==1) fprintf(stderr, "End ShellFish_HTTP_Update:system() Processing return data.\n");
	

	return New_Out_State;
}

/*
 * ShellFish_Run_Shell: Shell to event script and process return commands
 *
 * Return commands, each separated with a space. Ex. "P2 A127 B255 O6 o1"
 *
 *	O[n]	- Raise digital output. (O1..O8)
 *	o[n]	- Lower digital output. (o1..o8)
 *	P[n]	- Raise/Lower digital input (P1..P5)
 *	I[n]	- Raise digital input (I1..I5)
 *	i[n]	- Lower digital input (i1..i5)
 *	A[v]	- Set voltage analog out A (A0..A255)
 * 	B[v]	- Set voltage analog out B (B0..B255)
 * 	Y[v]	- Simulate voltage in A (Y0..Y255)
 * 	y[v]	- Set simulation in A (y0..y1)
 *      Z[v]	- Simulate voltage in B (Z0..Z255)
 * 	z[v]	- Set simulation in B (z0..z1)
 *	R[n]	- Reset counter (R1..R2)
 * 	T[n]	- Set timer (T1=secs..T255=secs)
 * 	F[n]	- Raise function event (F0..F255)
 */

int ShellFish_Run_Shell(int eventtype, int Current_Out_State, int Button_Number, int Current_Button_State, int Current_Delta_Step, 
                        int Counter_1, int *Sub_Counter_1, int Counter_2, int *Sub_Counter_2, int State_Analog_1, int State_Analog_2, 
			int *Dac_Out_State_1, int *Dac_Out_State_2, int *Queued_Button, int *Button_Hold, int *Function_Key, int *Timers,
			int *State_SimAnalogIn_1, int *State_SimAnalogIn_2, int *SimAnalogIn_1, int *SimAnalogIn_2)
{

	char Sys_Command[1024];			// Shell command
	int Sys_Return=0;			// Return value of shell
	int Shell_Return=0;			// Exit value of shell
	int i=0;				// Loop var
	char Response_Filename[1024];		// Filename of response file
	char Bit_String_Out_State[20];		// Bit formatted string of out state
	char Bit_String_Button_State[20];	// Bit formated string of button state
	
	char File_Char;				// char read from file
	char Command_Char=' ';			// char of command number
	int Command_Int=0;			// converted command number
	int New_Out_State = Current_Out_State;	// Copy current state
	
	// Set the analog in value to either actual or simulated input
	int StateA1in=State_Analog_1;				// actual input 1
	int StateA2in=State_Analog_2;				// actual input 2
	if(*SimAnalogIn_1==1) StateA1in=*State_SimAnalogIn_1;	// Change to analog simulated 1
	if(*SimAnalogIn_2==1) StateA2in=*State_SimAnalogIn_2;	// Change to analog simulated 2


	// Log message
	if(Logger_Mode==1) fprintf(stderr, "libk8055:WriteAllDigital(%i).\n",New_Out_State);
	
	// write output to usb hardware
	WriteAllDigital(New_Out_State);

	// copy the outstate
	strcpy(Bit_String_Out_State, Bit_Str(New_Out_State));
	
	// copy the buttate
	strcpy(Bit_String_Button_State, Bit_Str(Current_Button_State));
	
	// Add sub counters to counters
	int Cntr1=0;
	Cntr1=Counter_1 + *Sub_Counter_1;
	//if(Cntr1==-1) Cntr1=0;
	
	int Cntr2=0;
	Cntr2=Counter_2 + *Sub_Counter_2;
	//if(Cntr2==-1) Cntr2=0;
	
	// set command 
	if(eventtype== EV_BUTTON_SIG_DOWN){
		
		if(Quiet_Mode==0) printf("Input %i DN.\n",Button_Number);	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_BUTTON_SIG_DOWN_NAME, 
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
	
	}else if(eventtype== EV_BUTTON_SIG_DOWN_AFTER_HOLD){
		
		if(Quiet_Mode==0) printf("Input %i DH.\n",Button_Number);	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_BUTTON_SIG_DOWN_AFTER_HOLD_NAME, 
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
		
	}else if(eventtype== EV_BUTTON_SIG_UP){
		
		if(Quiet_Mode==0) printf("Input %i UP.\n",Button_Number);	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_BUTTON_SIG_UP_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2,  StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
		
	}else if(eventtype== EV_APP_START){
		
		if(Quiet_Mode==0) printf("Event ST.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_APP_START_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
	
	}else if(eventtype==EV_APP_EXIT){
		
		if(Quiet_Mode==0) printf("Event EX.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_APP_EXIT_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
	
	}else if(eventtype==EV_APP_REFRESH){
		
		if(Quiet_Mode==0) printf("Event RF.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_APP_REFRESH_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
	
	}else if(eventtype==EV_CARD_DISCONNECT){
		
		if(Quiet_Mode==0) printf("Event CD.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_CARD_DISCONNECT_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);
	
	}else if(eventtype==EV_CARD_CONNECT){
		
		if(Quiet_Mode==0) printf("Event CC.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_CARD_CONNECT_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_ANALOG1_VAL_CHANGED){
		
		if(Quiet_Mode==0) printf("Event A1.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_ANALOG1_VAL_CHANGED_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_ANALOG2_VAL_CHANGED){
		
		if(Quiet_Mode==0) printf("Event A2.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_ANALOG2_VAL_CHANGED_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_COUNTER1_VAL_CHANGED){
		
		if(Quiet_Mode==0) printf("Event C1.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_COUNTER1_VAL_CHANGED_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_COUNTER2_VAL_CHANGED){
		
		if(Quiet_Mode==0) printf("Event C2.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_COUNTER2_VAL_CHANGED_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_KEYBOARD_INPUT){
		
		if(Quiet_Mode==0) printf("Event KB.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_KEYBOARD_INPUT_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_TIMER){
		
		if(Quiet_Mode==0) printf("Event TM.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_TIMER_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}else if(eventtype==EV_FUNCTIONKEY){
		
		if(Quiet_Mode==0) printf("Event FN.\n");	

		snprintf(Sys_Command, 1024, "%s %i %i %i %s %s %s %i %i %i %i %i %i %i %i %i", 
		Run_File, Mod_Number, Usb_Board_Address, Button_Number, EV_FUNCTIONKEY_NAME,
		Bit_String_Out_State, Bit_String_Button_State, Current_Delta_Step, 
		Cntr1, Cntr2, StateA1in, StateA2in, *Dac_Out_State_1, *Dac_Out_State_2, *SimAnalogIn_1, *SimAnalogIn_2);

	}


	// Log message
	if(Logger_Mode==1) fprintf(stderr, "Start ShellFish_Run_Shell:system(%s)\n", Sys_Command);


	// command execution
	Sys_Return=system(Sys_Command);

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "Start ShellFish_Run_Shell:WEXITSTATUS\n");


	// Get the shell exit code
	Shell_Return=WEXITSTATUS(Sys_Return);

	// Log message
	if(Logger_Mode==1) fprintf(stderr, "End ShellFish_Run_Shell:system(), WEXITSTATUS: %i\n", Shell_Return);
	
	
	// Get the response data if exit code non zero
	if (Shell_Return!=0){
		
		// create filename
		snprintf(Response_Filename, 1024, "%s/shellfish.%i.%i.ret", Shell_Data_Folder, Mod_Number, Usb_Board_Address);

		// Message
		if(Quiet_Mode==0) printf("Opening file: %s\n", Response_Filename);
		
		// Open the response file
		FILE *File_Handle = fopen(Response_Filename, "r");
		if(File_Handle != NULL) {
			// Endless loop
			while (-1){
				// Get one char
				File_Char = fgetc(File_Handle);
				// If eof, then break
				if (feof(File_Handle)) break ;
		
				// On command
				if (File_Char=='O'){
					// Get output number
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Convert char to num
					Command_Int=Command_Char-48;
					// If the number is between 1 and 8
					if((Command_Int>0) & (Command_Int<9)){
						// Set de outstate
						New_Out_State=Merge_Bit(New_Out_State,Command_Int,1);
						// Message on screen
						if(Quiet_Mode==0) printf("Set output %i to ON\n", Command_Int);									
						
					}
					
				// Off command
				}else if (File_Char=='o'){
					// Get output number
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Convert char to num
					Command_Int=Command_Char-48;
					// If the number is 1 and 8
					if((Command_Int>0) & (Command_Int<9)){
						// Set de outstate
						New_Out_State=Merge_Bit(New_Out_State,Command_Int,0);
						// Message on screen
						if(Quiet_Mode==0) printf("Set output %i to OFF\n", Command_Int);
						
					}

				// Reset counter
				}else if (File_Char=='R'){
					
					// Get output number
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Convert char to num
					Command_Int=Command_Char-48;
					// If the number is between 1 and 8
					if((Command_Int>0) & (Command_Int<3)){
						
						// Log message
						if(Logger_Mode==1) fprintf(stderr, "libk8055:ResetCounter(%i).\n", Command_Int);
						
						// Reset counter
						ResetCounter(Command_Int);
						*Sub_Counter_1=0;
						*Sub_Counter_2=0;
						// Message on screen
						if(Quiet_Mode==0) printf("Reset counter %i\n", Command_Int);	
					}

				// Analog A
				}else if (File_Char=='A'){
					// Value of new dac
					int Analog_Value_A;
					// Get the new value from file stream
					Analog_Value_A=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Analog_Value_A>-1) & (Analog_Value_A<256)){
						// Set the parameter
						*Dac_Out_State_1= Analog_Value_A;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Analog out A to %i\n", Analog_Value_A);
					}
					
				// Analog B
				}else if (File_Char=='B'){
					// Value of new dac
					int Analog_Value_B;
					// Get the new value from file stream
					Analog_Value_B=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Analog_Value_B>-1) & (Analog_Value_B<256)){
						// Set the parameter
						*Dac_Out_State_2= Analog_Value_B;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Analog out B to %i\n", Analog_Value_B);
					}

				// Sim Analog in Y (1) in VALUE
				}else if (File_Char=='Y'){
					// Value of new dac
					int Analog_Value_Y;
					// Get the new value from file stream
					Analog_Value_Y=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Analog_Value_Y>-1) & (Analog_Value_Y<256)){
						// Turno on sim
						*SimAnalogIn_1=1;
						// Set the parameter
						*State_SimAnalogIn_1= Analog_Value_Y;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Simulated Analog in Y to %i\n", Analog_Value_Y);
					}

				// Sim Analog in y (1) on/off
				}else if (File_Char=='y'){
					// Value of new dac
					int Analog_Value_Y;
					// Get the new value from file stream
					Analog_Value_Y=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Analog_Value_Y>-1) & (Analog_Value_Y<2)){
						// Set the parameter
						*SimAnalogIn_1=Analog_Value_Y;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Simulation Analog in Y to %i\n", Analog_Value_Y);
					}


				// Sim Analog in Z (2) VALUE
				}else if (File_Char=='Z'){
					// Value of new dac
					int Analog_Value_Z;
					// Get the new value from file stream
					Analog_Value_Z=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Analog_Value_Z>-1) & (Analog_Value_Z<256)){
						// Turno on sim
						*SimAnalogIn_2=1;
						// Set the parameter
						*State_SimAnalogIn_2= Analog_Value_Z;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Simulated Analog in Z to %i\n", Analog_Value_Z);
					}
					
				// Sim Analog in z (2) on/off
				}else if (File_Char=='z'){
					// Value of new dac
					int Analog_Value_Z;
					// Get the new value from file stream
					Analog_Value_Z=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Analog_Value_Z>-1) & (Analog_Value_Z<2)){
						// Set the parameter
						*SimAnalogIn_2=Analog_Value_Z;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Simulation Analog in Z to %i\n", Analog_Value_Z);
					}

				// Button push/release command (emulate)
				}else if (File_Char=='P'){
					
					// Get output number
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Convert char to num
					Command_Int=Command_Char-48;
					// If the number is between 1 and 8
					if((Command_Int>-1) & (Command_Int<9)){
						// If all command
						if(Command_Int==0){
							// press/release all inputs
							for(i=1; i<6; i++){
								Queued_Button[i]=1;
								Button_Hold[i]=0;
							}
						}else{
							// Set de out / hold state
							Queued_Button[Command_Int]=1;
							Button_Hold[Command_Int]=0;
						}
						// Message on screen
						if(Quiet_Mode==0) printf("Emulate button %i Push/Release\n", Command_Int);									
						
					}

				// Button push command (emulate)
				}else if (File_Char=='I'){
					
					// Get output number
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Convert char to num
					Command_Int=Command_Char-48;
					// If the number is between 1 and 8
					if((Command_Int>0) & (Command_Int<9)){
						// Set de outstate
						Queued_Button[Command_Int]=1;
						Button_Hold[Command_Int]=1;
						// Message on screen
						if(Quiet_Mode==0) printf("Emulate button %i Push/Release\n", Command_Int);									
						
					}

				// Button release command (emulate)
				}else if (File_Char=='i'){
					
					// Get output number
					Command_Char = fgetc(File_Handle);
					// If eof, then break
					if (feof(File_Handle)) break ;
					// Convert char to num
					Command_Int=Command_Char-48;
					// If the number is between 1 and 8
					if((Command_Int>0) & (Command_Int<9)){
						// Set de outstate
						Queued_Button[Command_Int]=1;
						Button_Hold[Command_Int]=0;
						// Message on screen
						if(Quiet_Mode==0) printf("Emulate button %i Push/Release\n", Command_Int);	
					}
					
				}else if (File_Char=='F'){
					// Value of new dac
					int Function_Key_Value;
					// Get the new value from file stream
					Function_Key_Value=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Function_Key_Value>0) & (Function_Key_Value<256)){
						// Set the parameter
						Function_Key[Function_Key_Value]=1;
						// Message on screen
						if(Quiet_Mode==0) printf("Set Function call to %i\n", Function_Key_Value);
					}
					
				}else if (File_Char=='T'){
					// Value of new dac
					int Timer_Offset;
					int Timer_Value;
					// Get the new value from file stream
					Timer_Offset=fget_Numeric_Parm(File_Handle);
					// If value is 0-255
					if((Timer_Offset>0) & (Timer_Offset<256)){
						// If eof, then break
						if (feof(File_Handle)) break ;
						// Get number of seconds
						Timer_Value=fget_Numeric_Parm(File_Handle);
						// Eq or less than 24 Hrs
						if((Timer_Value>-1) & (Timer_Value<1000000)){
							// Set the parameters
							Timers[Timer_Offset]=Timer_Value*2;
							// Message on screen
							if(Quiet_Mode==0) printf("Set Timer %i to %i\n", Timer_Offset, Timer_Value);
						}
					}
				}
			}
			
			fclose(File_Handle);
			
		}else{
			
			if(Quiet_Mode==0) printf("Error reading: %s\n", Response_Filename);
			
			if(Logger_Mode==1) fprintf(stderr, "Error reading: %s\n", Response_Filename);
		
		}

	}

	return New_Out_State;

}

/*
 * File reads a maximum of 6 chars till eof or non-numeric and convert it to int (0-999999)
 * 
 */

int fget_Numeric_Parm(FILE *File_Handle)
{
	
	int Digit1=0;
	int Digit2=0;
	int Digit3=0;
	int Digit4=0;
	int Digit5=0;
	int Digit6=0;
	int Value=0;
	int Size=0;
	char Command_Char=' ';			// char of command number
	int Command_Int=0;			// converted command number
	
		
	while (-1){
		
		// Get one char 
		Command_Char = fgetc(File_Handle);
		// If eof, then break
		if (feof(File_Handle)) break ;
		// Convert char to num
		Command_Int=Command_Char-48;
		// If the number is between 1 and 8
		if((Command_Int>-1) & (Command_Int<10)){
			// Set the digit
			Digit1=Command_Int;
			Size++;
		}else{
			break;
		}
		
		// Get one char
		Command_Char = fgetc(File_Handle);
		// If eof, then break
		if (feof(File_Handle)) break ;
		// Convert char to num
		Command_Int=Command_Char-48;
		// If the number is between 1 and 8
		if((Command_Int>-1) & (Command_Int<10)){
			// Set the digit
			Digit2=Command_Int;
			Size++;
		}else{
			break;
		}

		// Get one char
		Command_Char = fgetc(File_Handle);
		// If eof, then break
		if (feof(File_Handle)) break ;
		// Convert char to num
		Command_Int=Command_Char-48;
		// If the number is between 1 and 8
		if((Command_Int>-1) & (Command_Int<10)){
			// Set the digit
			Digit3=Command_Int;
			Size++;
		}else{
			break;
		}
		
		// Get one char
		Command_Char = fgetc(File_Handle);
		// If eof, then break
		if (feof(File_Handle)) break ;
		// Convert char to num
		Command_Int=Command_Char-48;
		// If the number is between 1 and 8
		if((Command_Int>-1) & (Command_Int<10)){
			// Set the digit
			Digit4=Command_Int;
			Size++;
		}else{
			break;
		}

		// Get one char
		Command_Char = fgetc(File_Handle);
		// If eof, then break
		if (feof(File_Handle)) break ;
		// Convert char to num
		Command_Int=Command_Char-48;
		// If the number is between 1 and 8
		if((Command_Int>-1) & (Command_Int<10)){
			// Set the digit
			Digit5=Command_Int;
			Size++;
		}else{
			break;
		}
		
		// Get one char
		Command_Char = fgetc(File_Handle);
		// If eof, then break
		if (feof(File_Handle)) break ;
		// Convert char to num
		Command_Int=Command_Char-48;
		// If the number is between 1 and 8
		if((Command_Int>-1) & (Command_Int<10)){
			// Set the digit
			Digit6=Command_Int;
			Size++;
		}else{
			break;
		}

		break;

	}

	// Calculate the value based on the size of the number of digits
	if(Size==6){
		// Make a value of the 3 digits
		Value=(Digit1*100000)+(Digit2*10000)+(Digit3*1000)+(Digit4*100)+(Digit5*10)+Digit6;
	}else if(Size==5){
		// Make a value of the 3 digits
		Value=(Digit1*10000)+(Digit2*1000)+(Digit3*100)+(Digit4*10)+Digit5;
	}else if(Size==4){
		// Make a value of the 3 digits
		Value=(Digit1*1000)+(Digit2*100)+(Digit3*10)+Digit4;
	}else if(Size==3){
		// Make a value of the 3 digits
		Value=(Digit1*100)+(Digit2*10)+Digit3;
	}else if(Size==2){
		// Make a value of the 2 digits
		Value=(Digit1*10)+Digit2;
	}else{
		// Make a value of the 1 digits
		Value=Digit1;
	}
	
	return Value;
		
}


/* 
 * Bit_Str - Convert int value (0..255) to a bit formated char
 *
 * int is used as input, however, conversion does not exceed FF
 * Returns a char like "10101000"
 *
 */

const char *Bit_Str(int Byte_Value)
{
	// Byte_Value : Byte to convert to formatted bit string

	int bitpos[8];			// Array holding bits 
	int bp=1;			// Value of bit in byte 1,2,4,8, ...
	int i=0;			// Loop var
	static char text[9];		// bit string representation of the input value

	// Map the bits
        for (i=1;i<9;i++) {						// Cycle bits
		bitpos[i-1]=0;						// Init
		if (((Byte_Value & bp) == bp) == 1) bitpos[i-1]=1;	// If bit on, set it
		bp=bp*2;						// TO the next bit
 	}
	
	// Output formatted string
	snprintf(text, sizeof(text), "%i%i%i%i%i%i%i%i", bitpos[0],bitpos[1],bitpos[2],bitpos[3],bitpos[4],bitpos[5],bitpos[6],bitpos[7]);	
	
	// Return the formatted string
	return text;
}

/*
 * Merge_Bit: Merge a bit into existing offset and return the new value
 *
 */

int Merge_Bit(int Byte_Value, int Bit_Offset, int New_Bit_Value)
{
	// Byte_Value : Value where bit should be merged (0..255)
	// Bit_Offset : Position of bit (1..8)
	// Bit_Value  : Value of individual bit to replace with (1 or 0)
	
	int Ret_Byte=0;			// Return value
	int Bit_Value=0;		// Value of individual bit (1 or 0)
	int i=0;			// Loop var
	int bp=1;			// Value of bit in byte 1,2,4,8, ...

        for (i=1;i<9;i++) {						// Cycle bits
		Bit_Value=0;						// Initial value
		if (((Byte_Value & bp) == bp) == 1) Bit_Value=1;	// If bit is set
		if (i==Bit_Offset) Bit_Value=New_Bit_Value;		// If bit offset then use new bit value
		if (Bit_Value==1) Ret_Byte+=bp;				// Calculate byte value so far
		bp=bp*2;						// TO the next bit
 	}

	return Ret_Byte;		// Return the result
}

/*
 * Keyboard_Hit: Test if keyboard was hit
 *
 */

int Keyboard_Hit(void)
{
	struct timeval tv;
	fd_set read_fd;

	tv.tv_sec=0;
	tv.tv_usec=0;
	FD_ZERO(&read_fd);
	FD_SET(0,&read_fd);

	if(select(1, &read_fd, NULL, NULL, &tv) == -1)
		return 0;

	if(FD_ISSET(0,&read_fd))
		return 1;

	return 0;
}

/*
 * Get_Keyboard: Get input from keyboard
 *
 */

int Get_Keyboard(void)
{
    struct termios oldattr, newattr;
    int ch;
    tcgetattr( STDIN_FILENO, &oldattr );
    newattr = oldattr;
    newattr.c_lflag &= ~( ICANON | ECHO );
    tcsetattr( STDIN_FILENO, TCSANOW, &newattr );
    ch = getchar();
    tcsetattr( STDIN_FILENO, TCSANOW, &oldattr );
    return ch;
}

/*
 * Cv_String_To_Int: convert string to int
 *
 */

int Cv_String_To_Int(char *text, int *i)
{
	return sscanf(text, "%d", i);
}
